# -*- coding: utf-8 -*-
'''
Created on Fri Feb 12 19:00:29 2016

@author: Daniel Lefever
@email: lefeverde@gmail.com


This program was written using Biopython:

Cock, P.J.A. et al. Biopython: freely available Python tools for computational
molecular biology and bioinformatics. Bioinformatics 2009 Jun 1; 25(11) 1422-3
http://dx.doi.org/10.1093/bioinformatics/btp163 pmid:19304878

'''
#!/usr/bin/python

## These things make the program go ##


import sys
import argparse
import time
import os
import gc
import tempfile
import gzip
#import util_functions_dist

print('Working version')

from . import util_functions_dist

from .biopython import FastqGeneralIterator
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from time import sleep

start_time = time.time()


## This does the command line parsing ##
## Probably should include an example input ##

parser = argparse.ArgumentParser(add_help=False, description='This program splits \
combinatorially multiplexed paired-end seqs and outputs a folder for the R1 reads and \
another for the R2 reads. The two folders contain one fastq file per sample id.',
 usage='\npe_demuxer\
 -r1 r1_seqs.fastq -r2 r2_seqs.fastq -r1_bc r1_bc_file.txt -r2_bc -r2_bc_file.txt') #,add_help=False
req = parser.add_argument_group('required arguments')
opt = parser.add_argument_group('optional arguments')

req.add_argument('-r1_bc',
    help = 'BC file for R1 reads',
    metavar = '',
    required=True)
req.add_argument('-r2_bc',
    help = 'BC file for R2 reads',
    metavar = '' ,
    required=True)
req.add_argument('-r1',
    help = 'R1 fastq file',
    metavar = '',
    required=True)
req.add_argument('-r2',
    help = 'R2 fastq file',
    metavar = '',
    required=True)
#parser.add_argument('-h', '--help', action='store_true')
opt.add_argument("-h", "--help", action="help", help="show this help message and exit")
opt.add_argument('-o',
    help = 'name of output file or dir',
    default = 'pe_demuxer_output',
    metavar = '')
opt.add_argument('--batch_length',
    type=int,
    default = 1500000,
    help='amount of seqs to be put in a batch \
    only used when check_headers is passed (Default: 1500000)',
    metavar='')
opt.add_argument('--forward_primer_len',
    help = 'base number of primer/adapter on R1 seq (default: 0)',
    type=int,
    default=0,
    metavar = '')

opt.add_argument('--reverse_primer_len',
    help = 'base number of primer/adapter on R2 seq (default: 0)',
    type = int,
    default = 0,
    metavar = '')
opt.add_argument('--keep_original_headers',
    help = 'pass this arg to keep original headers in demultiplexed fasta/fastq file',
    action='store_true')
opt.add_argument('--min_seq_length',
    help = 'minimum length of seq to be retained (default: 20)',
    metavar = '',
    default = 20)
opt.add_argument('--check_headers',
    help = 'ensures the R1 and R2 read match before demultiplexing WARNING: this is super slow',
    action = 'store_true')


args = parser.parse_args()

example_usage = ('\n\t\t\tHow to use:\npe_demuxer\
 -r1 r1_seqs.fastq -r2 r2_seqs.fastq -r1_bc r1_bc_file.txt -r2_bc -r2_bc_file.txt')
example_r1_bc = ('\n\nThe -r1_bc argument specifies the barcode file corresponding to the r1 fastq file.\
The first column needs to correspond to letter (or whatever the front bc specifies)\
with the second being the barcode sequence.\
\nHere is an example of how this file should be formatted:' + \
'\n\tA GGTAC \n' + '\tB CAACAC \n' + '\tC ATCGGTT \n' + '---Note: the easiest way to make this file is\
to use excel and save it as a tab-delineated text file---\
\nAn example is provided by the example_data/forward_bcs.txt file.\n')
example_r2_bc = ('\nThe -r2_bc argument specifies the barcode file corresponding to the r2 fastq file.\
it should be formatted like the r1_bc file.\n')
example_r1_file = ('\nThe -r1 argument specifies the R1 seqs file.\
 See the example provided in the example_data dir for more information\n')
example_r2_file = ('\nThe -r2 argument specifies the R1 seqs file.\
 See the example provided in the example_data dir for more information\n')
example_keep = ('\n\nI dont actually think the --keep_original_headers\
does anything in this script, but it was too much trouble to remove.\n')
example_min = ('\nThe -min argument specifies what length the seq needs to be so that it is\
retained.\nDefault: 20')
example_check_header = ('\n\nThe --check_headers can handle situations where the seqs in the R1\
and R2 file are out of order and with reads missing from file.\nWARNING: this is going to be slow\
theres not much I can do about that, so I would recommend having your seqs in the same order.')
example_fp = '\n\nThe -fp argument specifies the how many bases after the barcode should be removed.\
The point of this is to remove primer/adapter sequences from reads without having to use any\
pattern matching. Thus, if you know how long the front primer/adapter sequences, you can just remove that\
many bases from the front of the read. If no value is passed, it defaults to 0.\n'
example_bp = '\n\nThe -bp argument is the same as the -fp, except it truncates so many bases from the end.\
Again, the point of this is to remove primer/adapter sequences without doing any sort of matching. This one \
also defaults to 0.\n'
example_output = ('\n\nThe -o just specifies the top dirs name. Two dirs within this dir\
 will be created, one for the R1 reads, and the other for the R2 reads.')



########################################################
''' These are the commands to make it go '''
########################################################

class PEDemux:
    def __init__(self):
        '''
        The PEDemux class takes care of Paired-end seqs.
        It demultiplexes by getting the front bits of off
        the R1 and R2 read. If the reads are in the same
        order and there is an R1 read for every R2 read,
        the alt_pe_loop can be used. Else, use the slow_pe_loop
        which will only get the overlapping set of reads determined
        by header. This is unfortunately much slower.

        '''
        self.args = args
        self.util_functions_dist = util_functions_dist
        self.r1_bc_dict, self.r1_len = self.util_functions_dist.bc_dict_maker(self.args.r1_bc, False)
        self.r2_bc_dict, self.r2_len = self.util_functions_dist.bc_dict_maker(self.args.r2_bc, False)
        self.sam_ind = self.util_functions_dist.sample_index_maker(self.r1_bc_dict, self.r2_bc_dict)
        self.seq_number = 0
        if self.args.r1.split('.')[-1] == 'gz':
            fm = 'wb'
        else:
            fm = 'w+'

        self.r1_file_dict = self.util_functions_dist.outfile_opener(
            self.args.o,
            self.sam_ind,
            file_mode=fm,
            dir_opt='R1')

        self.r2_file_dict = self.util_functions_dist.outfile_opener(
            self.args.o,
            self.sam_ind,
            file_mode=fm,
            dir_opt='R2'
            )


    # This creates a dict w/ with the seqs and a set w/ #
    # header info #
    def read_set_maker(self, seq_batch):
        '''
        Returns a dict with the header as key and the seq_rec
        tuple as item. It also returns a set of headers which
        for determining which overlap.
        '''
        read_dict = {}
        cur_set = set()
        for seq_rec in seq_batch:
            hd, seq, qual = seq_rec
            try:
                head_id, pe_crap = hd.split()
                cur_set.add(head_id)
                read_dict[head_id] = seq_rec
            except ValueError:
                cur_set.add(hd)
                read_dict[hd] = seq_rec
        return(read_dict, cur_set)

    # function which returns demultiplexed seqs #
    def pe_demultiplexed_seqrec_maker(self, r1_seq_rec, r2_seq_rec):
        '''
        This basically returns a tuple with the sample_id
        and both seq_recs. Fixed the nasty little code duplication
        from earlier.

        Possible stuff to be added in future:
        -fuzzy matching

        '''
        ## Borrowed Functions ##
        de_bc_loop = self.util_functions_dist.de_bc_loop
        demult_header = self.util_functions_dist.demult_header
                ##

        r1_head, r1_seq, r1_qual = r1_seq_rec
        r2_head, r2_seq, r2_qual = r2_seq_rec
        r1_front_bit = r1_seq[0:self.r1_len]
        r2_front_bit = r2_seq[0:self.r2_len]

        r1_comb, r1_start = de_bc_loop(r1_front_bit, self.r1_bc_dict)
        r2_comb, r2_start = de_bc_loop(r2_front_bit, self.r2_bc_dict)

        cur_r1_seq = r1_seq[(r1_start + self.args.forward_primer_len):len(r1_seq)]
        cur_r1_qual = r1_qual[(r1_start + self.args.forward_primer_len):len(r1_seq)]
        cur_r2_seq = r2_seq[(r2_start + self.args.reverse_primer_len):len(r1_seq)]
        cur_r2_qual = r2_qual[(r2_start + self.args.reverse_primer_len):len(r1_seq)]

        sample_id, header = demult_header(r1_comb, r2_comb, self.seq_number, \
            r1_head, self.args.keep_original_headers)


        r1_demux_seqrec = r1_head, cur_r1_seq, cur_r1_qual
        r2_demux_seqrec = r2_head, cur_r2_seq, cur_r2_qual

        return (sample_id, r1_demux_seqrec, r2_demux_seqrec)



    # Alternate demuxing loop #
    def alt_pe_loop(self, r1_batch, r2_batch):
        '''
        Basically does the normal demuxing with the assumption
        that the seqs in both files have a 1 to 1 correspondance
        and are in the same order. No intermediary steps are required
        as all 192 files are kept open.

        '''
        ### Borrowed Functions ####
        de_bc_loop = self.util_functions_dist.de_bc_loop
        demult_header = self.util_functions_dist.demult_header
        sample_index_maker = self.util_functions_dist.sample_index_maker
                    ###

        for (r1_cur_rec, r2_cur_rec) in zip(r1_batch, r2_batch):
            self.seq_number = self.seq_number + 1
            if self.seq_number % 1000 == 0:
                sys.stdout.write('Demultiplexed Seqs: ' '{0}\r'.format(self.seq_number),)
                sys.stdout.flush()

            sample_id, r1_demux_seqrec, r2_demux_seqrec = self.pe_demultiplexed_seqrec_maker(
                r1_cur_rec,
                r2_cur_rec)

            r1_head, cur_r1_seq, cur_r1_qual = r1_demux_seqrec
            r2_head, cur_r2_seq, cur_r2_qual = r2_demux_seqrec

            if (len(cur_r1_seq) < self.args.min_seq_length) or (len(cur_r2_seq) < self.args.min_seq_length):
                continue
            else:

                r1_cur_fastq = self.r1_file_dict[sample_id]
                r1_cur_fastq.write('@' + r1_head + \
                    '\n' + cur_r1_seq + '\n' + '+' +'\n' + cur_r1_qual + '\n')
                r2_cur_fastq = self.r2_file_dict[sample_id]
                r2_cur_fastq.write('@' + r2_head + \
                    '\n' + cur_r2_seq + '\n' + '+' +'\n' + cur_r2_qual + '\n')


    # main demuxing loop
    def slow_pe_loop(self, r1_seqdict, r1_set, r2_seqdict, r2_set):
        '''
        Basically does the same as the above function, but uses a set of
        headers from both files so that only reads with headers in both
        will be used.

        Really need to fix the duplication problem

        '''
        ### Borrowed Functions ###
        sample_index_maker = self.util_functions_dist.sample_index_maker
                ###

        pe_set = r1_set & r2_set
        pe_list = list(pe_set)
        r1_reads_dict = sample_index_maker(self.r1_bc_dict, self.r2_bc_dict)
        r2_reads_dict = sample_index_maker(self.r1_bc_dict, self.r2_bc_dict)

        for seq_id in pe_list:
            self.seq_number = self.seq_number + 1
            if self.seq_number % 1000 == 0:
                sys.stdout.write('Demultiplexed Seqs: ' '{0}\r'.format(self.seq_number),)
                sys.stdout.flush()

            r1_cur_rec = r1_seqdict[seq_id]
            r2_cur_rec = r2_seqdict[seq_id]

            sample_id, r1_demux_seqrec, r2_demux_seqrec =\
            pe_demultiplexed_seqrec_maker(r1_cur_rec, r2_cur_rec)

            r1_reads_dict[sample_id].append(r1_proc_seq)
            r2_reads_dict[sample_id].append(r2_proc_seq)

        return(r1_reads_dict, r2_reads_dict)


    def main_loop(self):
        ## Borrowed functions ##
        ind_fastq_writer = self.util_functions_dist.ind_fastq_writer
                        ##
        if self.args.r1.split('.')[-1] == 'gz':
            r1_iter = FastqGeneralIterator(gzip.open(self.args.r1, 'rU'))
            r2_iter = FastqGeneralIterator(gzip.open(self.args.r2, 'rU'))
        if self.args.r1.split('.')[-1] == 'fastq':
            r1_iter = FastqGeneralIterator(open(self.args.r1, 'rU'))
            r2_iter = FastqGeneralIterator(open(self.args.r2, 'rU'))


        if self.args.check_headers is True:
            r1_generator = self.util_functions_dist.batch_iterator(
                r1_iter,
                self.args.batch_length)
            r2_generator = self.util_functions_dist.batch_iterator(
                r2_iter,
                self.args.batch_length)
            for (r1_batch, r2_batch) in zip(r1_generator, r2_generator):
                cur_r1_dict, cur_r1_set = self.read_set_maker(r1_batch)
                cur_r2_dict, cur_r2_set = self.read_set_maker(r2_batch)
                r1_demux_dict, r2_demux_dict = self.slow_pe_loop(
                    cur_r1_dict,
                    cur_r1_set,
                    cur_r2_dict,
                    cur_r2_set)

                ind_fastq_writer(r1_demux_dict, self.r1_path, 'R1')
                ind_fastq_writer(r2_demux_dict, self.r2_path, 'R2')



        if self.args.check_headers is False:
            self.alt_pe_loop(
                r1_iter,
                r2_iter
                )
            for key, r1_file in self.r1_file_dict.items():
                r1_file.close()

            for key, r2_file in self.r2_file_dict.items():
                r2_file.close()




#if __name__ == '__main__':
#    pe_class = PEDemux()
#    pe_class.main_loop()
print(('\n' "Program Runtime: " + str(round(((time.time() - start_time))/60, 3)) + ' (min)' + '\n'))



#!/usr/bin/python


import mr_demuxy
from mr_demuxy import pe_demuxer_dist
#from mr_demuxy.pe_demuxer_dist import pe_demuxer_dist
#from .mr_demuxy import pe_demuxer_dist

def main():
    pe_class = pe_demuxer_dist.PEDemux()
    pe_class.main_loop()

if __name__ == '__main__':
    main()



#!/usr/bin/python



from Mr_Demuxy import merged_demuxer_dist

def main():
    merged_class = merged_demuxer_dist.MergedDemuxer()
    merged_class.main_loop()
    
if __name__ == '__main__':
    main()




#!/usr/bin/python

## These things make the program go ##


import sys
import os
import itertools
import gzip



from .biopython import FastqGeneralIterator

sys.dont_write_bytecode = True


#### USEFUL STUFF ####

def get_ambig_dna_values():
    '''
    Dict of IUPAC alphabets ambigious base calls
    '''
    ambiguous_dna_values = {
        "A": "A",
        "C": "C",
        "G": "G",
        "T": "T",
        "M": "AC",
        "R": "AG",
        "W": "AT",
        "S": "CG",
        "Y": "CT",
        "K": "GT",
        "V": "ACG",
        "H": "ACT",
        "D": "AGT",
        "B": "CGT",
        "X": "GATC",
        "N": "GATC",
        }
    return ambiguous_dna_values


def get_ambig_complement():
    '''
    Dict of IUPAC base call complement
    '''
    ambiguous_dna_complement = {
        "A": "T",
        "C": "G",
        "G": "C",
        "T": "A",
        "M": "K",
        "R": "Y",
        "W": "W",
        "S": "S",
        "Y": "R",
        "K": "M",
        "V": "B",
        "H": "D",
        "D": "H",
        "B": "V",
        "X": "X",
        "N": "N",
        }
    return ambiguous_dna_complement


# Line counter #

def file_len(fname):
    '''
taken from
http://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python?lq=1
    '''
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

# This just converts amig bases to bracketed bases. #
# Useful for making a 'grep'-able string #
def ambig_base_remover(in_seq):
    nuc_list = ['A', 'C', 'G', 'T']
    ambig_base_dict = get_ambig_dna_values()
    fixed_seq = ''
    for nuc in in_seq:
        if nuc in nuc_list:
            fixed_seq = fixed_seq + nuc
        if nuc not in nuc_list:
            actual_base = '[' + ambig_base_dict[nuc] + ']'
            fixed_seq = fixed_seq + actual_base
    return fixed_seq



# Simple reverse complementer #
def reverse_complement(seq):
    comp_dict = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}
    reversed_seq = seq[::-1]
    rev_comp_seq = ''
    for nuc in reversed_seq:
        rev_comp_seq = rev_comp_seq + comp_dict[nuc]
    return rev_comp_seq


# This generates  the files for appending #
def outfile_maker(files_path, sample_dict, dir_opt=None):
    if dir_opt is None:
        cur_dir = files_path
        fq_end = '.fastq'
    if dir_opt is not None:
        cur_dir = os.path.join(files_path, dir_opt)
        fq_end = '_' + dir_opt + '.fastq'
    if not os.path.exists(cur_dir):
        os.makedirs(cur_dir)
    for key in sample_dict.keys():
        cur_handle = os.path.join(cur_dir, (key + fq_end))
        with open(cur_handle, 'w+') as f:
            pass
    return cur_dir

# writes files and leaves open #
def outfile_opener(files_path, sample_dict, file_mode='w+', dir_opt=None):
    file_dict = {}
    if dir_opt is None:
        cur_dir = files_path
        fq_end = '.fastq'
    if dir_opt is not None:
        cur_dir = os.path.join(files_path, dir_opt)
        fq_end = '_' + dir_opt + '.fastq'
    if not os.path.exists(cur_dir):
        os.makedirs(cur_dir)
    for key in sample_dict.keys():
        cur_handle = os.path.join(cur_dir, (key + fq_end))
        if file_mode == 'w+':
            file_dict[key] = open(cur_handle, file_mode)
        if file_mode == 'wb':
            file_dict[key] = gzip.open((cur_handle + '.gz'), file_mode)
    return file_dict


# This makes the bc dicts #
def bc_dict_maker( bc_file, rev_comp_bcs):
    bc_dict = {}
    bc_len = []
    for line in open(bc_file, 'rU'):
        comb_id, bc = line.split()
        bc_len.append(len(bc))
        rev_bc = reverse_complement(bc)
        if rev_comp_bcs is True:
            bc_dict[comb_id] = rev_bc
        if rev_comp_bcs is False:
            bc_dict[comb_id] = bc
    return (bc_dict, max(bc_len))

# This just trims the read #
def trimmed_read(seq, qual, start_pos, end_pos, fp_len, bp_len, infile_extension):
    cur_seq = seq\
            [(start_pos + fp_len):(len(seq) - (end_pos + bp_len))]
    if infile_extension == 'fasta':
        cur_qual = None
        return (cur_seq, cur_qual)
    if infile_extension == 'fastq':
        cur_qual = qual\
            [(start_pos + fp_len):(len(seq) - (end_pos + bp_len))]
        return (cur_seq, cur_qual)


# Creates a dict with sampleID as key and
# list as item
def sample_index_maker(dict1, dict2):
    index_dict = {}
    sample_list = []
    for r1 in dict1.keys():
        for r2 in dict2.keys():
            sample_list.append(r1 + r2)
    for i in sample_list:
        index_dict[i] = []
        index_dict['Unassigned'] = []
    return index_dict

# this is what writes each of ind files by appending #
def ind_fastq_writer(cur_index, files_path, dir_opt=None):
    if dir_opt is None:
        cur_dir = files_path
        fq_end = '.fastq'
    if dir_opt is not None:
        cur_dir = files_path #+ '/' + dir_opt
        fq_end = '_' + dir_opt + '.fastq'
    for key, seq_list in cur_index.items():
        for seq_rec in seq_list:
            hd, seq, qual = seq_rec
            cur_handle = os.path.join(cur_dir, (key + fq_end))
            with open(cur_handle, 'a+') as f:
                f.write('@' + hd + '\n'+ seq + '\n+\n' + qual + '\n')

def ind_fastq_chunk_writer(cur_index, files_path, dir_opt=None):
    if dir_opt is None:
        cur_dir = files_path
        fq_end = '.fastq'
    if dir_opt is not None:
        cur_dir = files_path
        fq_end = '_' + dir_opt + '.fastq'
    for key, seq_chunk in cur_index.items():
        cur_handle = os.path.join(cur_dir, (key + fq_end))
        with open(cur_handle, 'a+') as f:
            f.write(str(seq_chunk))



def single_fasta_fastq(seq_rec, file_extension, out_file, out_qual_file, min_seq_length):

    header, cur_seq, cur_qual = seq_rec

    if file_extension == 'fasta':
        if len(cur_seq) < min_seq_length:
            return
        else:
            out_file.write('>' + header + '\n' + cur_seq + '\n')
            if out_qual_file is not None:
                out_qual_file.write('>' + header + '\n' + cur_qual + '\n')


    if file_extension == 'fastq':
        if len(cur_seq) < min_seq_length:
            return
        else:
            out_file.write('@' + header + '\n' + cur_seq \
            + '\n' + '+' +'\n' + cur_qual + '\n')




# Determines infile format #
def infile_extension(infile):
   with open(infile, 'rU') as f:
       first_char = f.read(1)
       if first_char == '@':
           return(str('fastq'))
       if first_char == '>':
           return(str('fasta'))

 # This creates a dict w/ with the seqs and a set w/ #
 # header info #
def read_set_maker( seq_batch):
    read_dict = {}
    cur_set = set()
    for seq_rec in seq_batch:
        hd, seq, qual = seq_rec
        try:
            head_id, pe_crap = hd.split()
            cur_set.add(head_id)
            read_dict[head_id] = seq_rec
        except ValueError:
            cur_set.add(hd)
            read_dict[hd] = seq_rec
    return(read_dict, cur_set)

def demult_header( r1_id, r2_id, seq_number, orig_header, keep_original_headers):
   if 'Unk' in (r1_id + r2_id):
       sample_id = 'Unassigned'
   if 'Unk' not in (r1_id + r2_id):
       sample_id = r1_id + r2_id
   if keep_original_headers is True:
       header = orig_header + ' ' + sample_id + '_' + str(seq_number)
   if keep_original_headers is False:
       header = sample_id + '_' + str(seq_number)
   return(sample_id, header)

 # This function takes the bc containing bits #
 # and matches them to the dicts #
def de_bc_loop( sub_seq, bc_dict):
    for key, bc in bc_dict.items():
        if bc in sub_seq:
            comb_id = key
            start_pos = len(bc)
            break
        if bc not in sub_seq:
            comb_id = 'Unk'
            start_pos = 0
    return (comb_id, start_pos)




# Batch Generator Function taken from Biopython website #
# http://biopython.org/wiki/Split_large_file #

def batch_iterator( iterator, batch_size) :
    """Returns lists of length batch_size.

    This can be used on any iterator, for example to batch up
    SeqRecord objects from Bio.SeqIO.parse(...), or to batch
    Alignment objects from Bio.AlignIO.parse(...), or simply
    lines from a file handle.

    This is a generator function, and it returns lists of the
    entries from the supplied iterator.  Each list will have
    batch_size entries, although the final list may be shorter.
    """
    entry = True #Make sure we loop once
    while entry :
        batch = []
        while len(batch) < batch_size :
            try :
                entry = next(iterator)
            except StopIteration :
                entry = None
            if entry is None :
                #End of file
                break
            batch.append(entry)
        if batch :
            yield batch

def mc_spinner(proc='\n'):
    sys.stdout.write(proc + '\n')
    sys.stdout.flush()
    spinner = itertools.cycle(['-', '/', '|', '\\'])
    while True:
        sys.stdout.write(next(spinner))
        sys.stdout.flush()
        time.sleep(0.1)
        sys.stdout.write('\b')



#!/usr/bin/python


import sys

sys.dont_write_bytecode = True

def FastqGeneralIterator(handle):
    """Iterate over Fastq records as string tuples (not as SeqRecord objects).

    This code does not try to interpret the quality string numerically.  It
    just returns tuples of the title, sequence and quality as strings.  For
    the sequence and quality, any whitespace (such as new lines) is removed.

    Our SeqRecord based FASTQ iterators call this function internally, and then
    turn the strings into a SeqRecord objects, mapping the quality string into
    a list of numerical scores.  If you want to do a custom quality mapping,
    then you might consider calling this function directly.

    For parsing FASTQ files, the title string from the "@" line at the start
    of each record can optionally be omitted on the "+" lines.  If it is
    repeated, it must be identical.

    The sequence string and the quality string can optionally be split over
    multiple lines, although several sources discourage this.  In comparison,
    for the FASTA file format line breaks between 60 and 80 characters are
    the norm.

    **WARNING** - Because the "@" character can appear in the quality string,
    this can cause problems as this is also the marker for the start of
    a new sequence.  In fact, the "+" sign can also appear as well.  Some
    sources recommended having no line breaks in the  quality to avoid this,
    but even that is not enough, consider this example::

        @071113_EAS56_0053:1:1:998:236
        TTTCTTGCCCCCATAGACTGAGACCTTCCCTAAATA
        +071113_EAS56_0053:1:1:998:236
        IIIIIIIIIIIIIIIIIIIIIIIIIIIIICII+III
        @071113_EAS56_0053:1:1:182:712
        ACCCAGCTAATTTTTGTATTTTTGTTAGAGACAGTG
        +
        @IIIIIIIIIIIIIIICDIIIII<%<6&-*).(*%+
        @071113_EAS56_0053:1:1:153:10
        TGTTCTGAAGGAAGGTGTGCGTGCGTGTGTGTGTGT
        +
        IIIIIIIIIIIICIIGIIIII>IAIIIE65I=II:6
        @071113_EAS56_0053:1:3:990:501
        TGGGAGGTTTTATGTGGA
        AAGCAGCAATGTACAAGA
        +
        IIIIIII.IIIIII1@44
        @-7.%<&+/$/%4(++(%

    This is four PHRED encoded FASTQ entries originally from an NCBI source
    (given the read length of 36, these are probably Solexa Illumina reads where
    the quality has been mapped onto the PHRED values).

    This example has been edited to illustrate some of the nasty things allowed
    in the FASTQ format.  Firstly, on the "+" lines most but not all of the
    (redundant) identifiers are omitted.  In real files it is likely that all or
    none of these extra identifiers will be present.

    Secondly, while the first three sequences have been shown without line
    breaks, the last has been split over multiple lines.  In real files any line
    breaks are likely to be consistent.

    Thirdly, some of the quality string lines start with an "@" character.  For
    the second record this is unavoidable.  However for the fourth sequence this
    only happens because its quality string is split over two lines.  A naive
    parser could wrongly treat any line starting with an "@" as the beginning of
    a new sequence!  This code copes with this possible ambiguity by keeping
    track of the length of the sequence which gives the expected length of the
    quality string.

    Using this tricky example file as input, this short bit of code demonstrates
    what this parsing function would return:

    >>> with open("Quality/tricky.fastq", "rU") as handle:
    ...     for (title, sequence, quality) in FastqGeneralIterator(handle):
    ...         print(title)
    ...         print("%s %s" % (sequence, quality))
    ...
    071113_EAS56_0053:1:1:998:236
    TTTCTTGCCCCCATAGACTGAGACCTTCCCTAAATA IIIIIIIIIIIIIIIIIIIIIIIIIIIIICII+III
    071113_EAS56_0053:1:1:182:712
    ACCCAGCTAATTTTTGTATTTTTGTTAGAGACAGTG @IIIIIIIIIIIIIIICDIIIII<%<6&-*).(*%+
    071113_EAS56_0053:1:1:153:10
    TGTTCTGAAGGAAGGTGTGCGTGCGTGTGTGTGTGT IIIIIIIIIIIICIIGIIIII>IAIIIE65I=II:6
    071113_EAS56_0053:1:3:990:501
    TGGGAGGTTTTATGTGGAAAGCAGCAATGTACAAGA IIIIIII.IIIIII1@44@-7.%<&+/$/%4(++(%

    Finally we note that some sources state that the quality string should
    start with "!" (which using the PHRED mapping means the first letter always
    has a quality score of zero).  This rather restrictive rule is not widely
    observed, so is therefore ignored here.  One plus point about this "!" rule
    is that (provided there are no line breaks in the quality sequence) it
    would prevent the above problem with the "@" character.
    """
    # We need to call handle.readline() at least four times per record,
    # so we'll save a property look up each time:
    handle_readline = handle.readline

    # Skip any text before the first record (e.g. blank lines, comments?)
    while True:
        line = handle_readline()
        if not line:
            return  # Premature end of file, or just empty?
        if line[0] == "@":
            break
        if isinstance(line[0], int):
            raise ValueError("Is this handle in binary mode not text mode?")

    while line:
        if line[0] != "@":
            raise ValueError(
                "Records in Fastq files should start with '@' character")
        title_line = line[1:].rstrip()
        # Will now be at least one line of quality data - in most FASTQ files
        # just one line! We therefore use string concatenation (if needed)
        # rather using than the "".join(...) trick just in case it is multiline:
        seq_string = handle_readline().rstrip()
        # There may now be more sequence lines, or the "+" quality marker line:
        while True:
            line = handle_readline()
            if not line:
                raise ValueError("End of file without quality information.")
            if line[0] == "+":
                # The title here is optional, but if present must match!
                second_title = line[1:].rstrip()
                if second_title and second_title != title_line:
                    raise ValueError("Sequence and quality captions differ.")
                break
            seq_string += line.rstrip()  # removes trailing newlines
        # This is going to slow things down a little, but assuming
        # this isn't allowed we should try and catch it here:
        if " " in seq_string or "\t" in seq_string:
            raise ValueError("Whitespace is not allowed in the sequence.")
        seq_len = len(seq_string)

        # Will now be at least one line of quality data...
        quality_string = handle_readline().rstrip()
        # There may now be more quality data, or another sequence, or EOF
        while True:
            line = handle_readline()
            if not line:
                break  # end of file
            if line[0] == "@":
                # This COULD be the start of a new sequence. However, it MAY just
                # be a line of quality data which starts with a "@" character.  We
                # should be able to check this by looking at the sequence length
                # and the amount of quality data found so far.
                if len(quality_string) >= seq_len:
                    # We expect it to be equal if this is the start of a new record.
                    # If the quality data is longer, we'll raise an error below.
                    break
                # Continue - its just some (more) quality data.
            quality_string += line.rstrip()

        if seq_len != len(quality_string):
            raise ValueError("Lengths of sequence and quality values differs "
                             " for %s (%i and %i)."
                             % (title_line, seq_len, len(quality_string)))

        # Return the record and then continue...
        yield (title_line, seq_string, quality_string)
    raise StopIteration




import os
import site
import subprocess
import sys


from distutils.core import setup

# Functions to help w/ install #
def file_path_joiner(file_dir):
    files_list = os.listdir(file_dir)
    corrected_file_path = []
    for cur_file in files_list:
        cur_fp = os.path.join(file_dir, cur_file)
        corrected_file_path.append(cur_fp)
    return corrected_file_path

def bash_checker(in_file, line_want):
    append_file = True
    for line in in_file:
        if line_want in line:
            append_file = False
            break

        if line_want not in line:
            append_file = True
    return append_file

# vars for making program work out of box #

script_files = file_path_joiner('bin')
user_home = os.path.expanduser('~')
user_path = site.USER_BASE
export_line = 'export PATH=' + user_path + '/bin:$PATH'
bash_file = os.path.join(user_home, '.bash_profile')


setup(
    name = 'Mr_Demuxy',
    version = '1.2.1',
    description = 'demultiplexs combinatorially tagged reads',

    author = 'Daniel E. Lefever',
    author_email = 'lefeverde@gmail.com',

    packages = ['Mr_Demuxy'],
    package_dir = {
    'Mr_Demuxy': 'mr_demuxy',
    },
    scripts = script_files,

    package_data = {
    'Mr_Demuxy': ['example_data/example*'],
    'Mr_Demuxy': ['example_data/*.txt']
    },
    data_files = [
    ('Mr_Demuxy', ['MANIFEST.in']),
    ('', ['LICENSE/biopython_license.txt'])],
    url = 'https://pypi.python.org/pypi/Mr_Demuxy/',

    license = 'MIT',
    python_requires='>=2.7, >=3.6, <4',
    platforms = 'any',
    classifiers = [
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        ]

    )

sys.stdout.write('\nensuring local dir is in path\n')
sys.stdout.flush()
try:
    sys.stdout.write('bash_profile found\nadding local dir to path\n')
    sys.stdout.flush()

    pe_alias = 'alias pe_demuxer'
    merg_alias = 'alias merged_demuxer'

    open_bf = open(bash_file, 'rU')
    bash_lines = open_bf.readlines()
    open_bf.close()

    add_local_path = bash_checker(bash_lines, export_line)

    with open(bash_file, 'w+') as f:
        if add_local_path is True:
            f.write(
                '\n#Added by Mr. Demuxy\n'
                + export_line + '\n\n'
                )
        for line in bash_lines:
            if pe_alias not in line:
                if merg_alias not in line:
                    f.write(line)


except IOError:
    sys.stdout.write('creating new bash_profile\nadding local dir to path\n')
    sys.stdout.flush()
    with open(bash_file, 'w+') as f:
        f.write(
            '\n#Added by Mr. Demuxy\n'
             + export_line + '\n'
             )



# -*- coding: utf-8 -*-
'''
Created on Fri Feb 12 19:00:29 2016

@author: Daniel Lefever
@email: lefeverde@gmail.com


This program was written using Biopython:

Cock, P.J.A. et al. Biopython: freely available Python tools for computational
molecular biology and bioinformatics. Bioinformatics 2009 Jun 1; 25(11) 1422-3
http://dx.doi.org/10.1093/bioinformatics/btp163 pmid:19304878

'''
#!/usr/bin/python


## These things make the program go ##



import sys
import argparse
import time
import os
import gc
import tempfile
from . import util_functions_dist


from .biopython import FastqGeneralIterator
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from time import sleep

start_time = time.time()


## This does the command line parsing ##
## Probably should include an example input ##

parser = argparse.ArgumentParser(add_help=False,
    description='This program demultiplexes combinatorially barcoded reads.',
    usage='\nmerged_demuxer -f forward_bcs.txt -r reverse_bcs.txt \
-i example_merged.fastq  -o demultiplexed_seqs.fasta') #, add_help=False
req = parser.add_argument_group('required arguments')
opt = parser.add_argument_group('optional arguments')
req.add_argument('-f',
     help='forward bcs file',
     metavar = '',
     required=True)
req.add_argument('-r',
    help = 'reverse bcs file',
    metavar = '',
    required=True)
req.add_argument('-i',
    help = 'input merged fastq file',
    metavar = '',
    required=True)
opt.add_argument('-o',
    help = 'name of output file or folder (default: merged_demuxer_out)',
    default = 'merged_demuxer_out',
    metavar = '')
opt.add_argument('--forward_primer_len',
    help = 'base number of forward primer/adapter (default: 0)',
    type=int,
    default=0,
    metavar = '')
opt.add_argument('--reverse_primer_len',
    help = 'base number of reverse primer/adapter (default: 0)',
    type = int,
    default = 0,
    metavar = '')

opt.add_argument('--no_rev_comp_reverse_bcs',
    action = 'store_false',
    help ='prevents rev bcs from reverse complementing'
    )
opt.add_argument("-h", "--help", action="help", help="show this help message and exit")

opt.add_argument('--keep_original_headers',
    help = 'pass this arg to keep original headers in demultiplexed fasta/fastq file',
    action='store_true')
opt.add_argument('--write_qual',
    action='store_true',
    help='write a qual file containing quality scores')
opt.add_argument('--individual_fastq',
    action='store_true',
    help='demultiplex by writing a fastq file for each sample ID')
opt.add_argument('--min_seq_length',
    help = 'minimum length of seq to be retained (default: 20)',
    metavar = '',
    default = 20)
opt.add_argument('--file_extension',
    default='fasta',
    help='format of the output file (default: fasta)',
    metavar = '')



example_usage = ('\n\t\t\tHere is an example of how to use this program:\
\n-f forward_bcs.txt -r reverse_bcs.txt \
-i example_merged.fastq  -o demultiplexed_seqs.fasta')
example_forward_bcs = '\nThe -f argument specifies the forward barcode file.\
The first column needs to correspond to letter (or whatever the front bc specifies)\
with the second being the barcode sequence.\
\nHere is an example of how this file should be formatted:' + \
'\n\tA GGTAC \n' + '\tB CAACAC \n' + '\tC ATCGGTT \n' + '---Note: the easiest way to make this file is\
to use excel and save it as a tab-delineated text file---\
\nAn example is provided by the example_data/forward_bcs.txt file.\n'
example_reverse_bcs = '\n\nThe -r argument specifies the reverse barcode file. like the forward_bcs file, \
it needs to have the first column to correspond to number (or whatever the back bc specifies) \
with the second corresponding to barcode sequence\
\nHere is example of how this one should be formatted:' + \
'\n1 AGGAA \n' + '2 GAGTGG\n' + '3 CCACGTC\n \
An example is provided by the example_data/reverse_bcs.txt file\n'
example_input_fasta = '\n\nThe -i argument specifies the input merged fastq file. This is the unmultiplexed file \
created by merging pair-ended fastq files. Note: this script only accepts merged fastq files.\
\nFor paired-end files, use the pe_demuxer script. \nThe merged fastq should looks somehting like this:\n\n\
@M02849:138:000000000-AJTC8:1....\nGGTACCCTACGG.....GGTAGTCTTCCT......'
example_output_fasta = '\n\nThe -o option just specifies what to the output file or dir will be called.\
 Call it whatever you want.\n'
example_fp = '\nThe -fp argument specifies the how many bases after the barcode should be removed.\
The point of this is to remove primer/adapter sequences from reads without having to use any\
pattern matching. Thus, if you know how long the front primer/adapter sequences, you can just remove that\
many bases from the front of the read. If no value is passed, it defaults to 0.\n'
example_bp = '\nThe -bp argument is the same as the -fp, except it truncates so many bases from the end.\
Again, the point of this is to remove primer/adapter sequences without doing any sort of matching. This one \
also defaults to 0.\n'
example_ext = '\nThe -ext argument specifies the output file type. The default setting is fasta. \
to get a fastq file, pass -ext fastq.'
example_ind = '\nThe -ind argument will split the fastq into individual fq files with one sample per file\n '
example_keep = '\nThe -keep argument tells the program whether to keep the orignal illumina fastq header \
information when a single fastq, fasta, or fasta and qual are the output.\
The new header becomes @<orignal_header_from_sequencers> <demultiplexed_seqID>'
example_write_qual = '\n\nThe --write_qual argument allows the quality scores of each seq to written into a \
seperate file. Simply pass the argument to use.'
args = parser.parse_args()







########################################################
''' These are the commands to make it go '''
########################################################

class MergedDemuxer:
    def __init__(self):
        '''
        Class for handling the demultiplexing of merged reads

        '''
        self.args = args
        self.util_functions = util_functions_dist
        self.fbc_dict, self.fbc_len = self.util_functions.bc_dict_maker(
            self.args.f,
            False)
        self.rbc_dict, self.rbc_len = self.util_functions.bc_dict_maker(
            self.args.r,
            self.args.no_rev_comp_reverse_bcs)
        self.sam_ind = self.util_functions.sample_index_maker(self.fbc_dict, self.rbc_dict)
        self.seq_number = 0
        self.infile_extension = self.util_functions.infile_extension(self.args.i)
        self.out_path = os.path.abspath(self.args.o)


    def demultiplexed_seqrec_maker(self, seq_rec):
        de_bc_loop = self.util_functions.de_bc_loop
        trimmed_read = self.util_functions.trimmed_read
        demult_header = self.util_functions.demult_header

        orig_header, seq, qual = seq_rec
        front_subseq = seq[0:self.fbc_len]
        back_subseq = seq[-self.rbc_len:]

        front_id, start_pos = de_bc_loop(front_subseq, self.fbc_dict)

        back_id, end_pos = de_bc_loop(back_subseq, self.rbc_dict)

        cur_seq, cur_qual = trimmed_read(seq, qual, start_pos, end_pos,\
            self.args.forward_primer_len, self.args.reverse_primer_len, self.infile_extension)

        sample_id, cur_header = demult_header(front_id, back_id, self.seq_number, orig_header,\
            self.args.keep_original_headers)

        demult_seq_rec = (sample_id, orig_header, cur_header, cur_seq, cur_qual)

        return demult_seq_rec


    def merged_demux_loop(self, batch, file_dict=None, out_file=None, out_qual=None):
        '''
        This is the main demuxing looping

        '''
        demultiplexed_seqrec_maker = self.demultiplexed_seqrec_maker

        for seq_rec in batch:
            self.seq_number = self.seq_number + 1
            if self.seq_number % 1000 == 0:
                sys.stdout.write('Demuxed Seqs: ' '{0}\r'.format(self.seq_number),)
                sys.stdout.flush()

            sample_id, orig_header, cur_header, cur_seq, cur_qual = demultiplexed_seqrec_maker(seq_rec)

            if self.args.individual_fastq is False:

                if self.args.file_extension == 'fasta':
                    if len(cur_seq) < self.args.min_seq_length:
                        continue
                    elif self.args.write_qual is False:
                        out_file.write('>' + cur_header + '\n' + cur_seq + '\n')
                        continue
                    else:
                        out_file.write('>' + cur_header + '\n' + cur_seq + '\n')
                        out_qual.write('>' + cur_header + '\n' + cur_qual + '\n')
                        continue

                if self.args.file_extension == 'fastq':
                    if len(cur_seq) < self.args.min_seq_length:
                        continue
                    else:
                        out_file.write('@' + cur_header + '\n' + cur_seq \
                        + '\n' + '+' +'\n' + cur_qual + '\n')
                        continue

            if self.args.individual_fastq is True:
                cur_fastq = file_dict[sample_id]
                if len(cur_seq) < self.args.min_seq_length:
                    continue
                else:
                    cur_fastq.write('@' + orig_header + \
                    '\n' + cur_seq + '\n' + '+' +'\n' + cur_qual + '\n')


        return file_dict

    def main_loop(self):
        '''
        This is going demultiplex and split everything all in one go,
        since the files are broken into bits anyhow
        '''
        fq_generator = FastqGeneralIterator(open(self.args.i))
        merged_demux_loop = self.merged_demux_loop


        if self.args.individual_fastq is True:
            file_dict = self.util_functions.outfile_opener(self.out_path, self.sam_ind)
            merged_demux_loop(
                fq_generator,
                file_dict = file_dict
                )
            for key, file in file_dict.items():
                file.close()


        if self.args.individual_fastq is False:
            if self.args.write_qual is True:
                out_file_open = open((self.out_path + '.' + self.args.file_extension), 'w+')
                out_qual_open = open((self.out_path + '.qual'), 'w+')
                merged_demux_loop(
                    fq_generator,
                    out_file=out_file_open,
                    out_qual=out_qual_open
                    )
                out_qual_open.close()
                out_file_open.close()

        if self.args.individual_fastq is False:
            if self.args.write_qual is False:
                out_file_open = open((self.out_path + '.' + self.args.file_extension), 'w+')
                merged_demux_loop(
                    fq_generator,
                    out_file=out_file_open
                    )
                out_file_open.close()


#if __name__ == '__main__':

#    sys.stdout.write('\n')
#    sys.stdout.flush()
#    merged_class = MergedDemuxer()
#    merged_class.main_loop()
#    sys.dont_write_bytecode = True

print(('\n' "Program Runtime: " + str(round(((time.time() - start_time))/60, 3)) + ' (min)' + '\n'))
